# Written by Cameron Dale
# see LICENSE.txt for license information
#
# $Id: HTTPCache.py 220 2007-08-11 21:07:36Z camrdale-guest $

"""Manage an HTTP download cache.

@type logger: C{logging.Logger}
@var logger: the logger to send all log messages to for this module
@type time_format: C{string}
@var time_format: the format to use for reading/writing HTTP server times
@type VERSION: C{string}
@var VERSION: the UserAgent identifier sent to all sites
@type alas: C{string}
@var alas: the message to send when the data is not found

"""

from httplib import HTTPConnection
from threading import Thread
from DebTorrent.__init__ import product_name,version_short
from os.path import join, split, getmtime, getsize, exists
from os import utime, makedirs, listdir
from time import strftime, strptime, gmtime
from calendar import timegm
import logging
try:
    True
except:
    True = 1
    False = 0

logger = logging.getLogger('DebTorrent.HTTPCache')

time_format = '%a, %d %b %Y %H:%M:%S'
VERSION = product_name+'/'+version_short
alas = 'your file may exist elsewhere in the universe\nbut alas, not here\n'

class CacheRequest:
    """Download a file needed for the HTTP download cache.
    
    @type handler: L{HTTPCache}
    @ivar handler: the cache manager for the download
    @type path: C{list} of C{string}
    @ivar path: the server and path to download
    @type server: C{string}
    @ivar server: the webserver address and port to connect to 
    @type url: C{string}
    @ivar url: the URL to request from the site
    @type func: C{method}
    @ivar func: the method to call when the download completes
    @type connection: C{HTTPConnection}
    @ivar connection: the connection to the HTTP server
    @type headers: C{dictionary}
    @ivar headres: the HTTP headers to send in the request, and the headers
        returned by the response
    @type active: C{boolean}
    @ivar active: whether there is a download underway
    @type received_data: C{string}
    @ivar received_data: the data returned from the server
    @type connection_status: C{int}
    @ivar connection_status: the status code returned by the server
    @type connection_response: C{string}
    @ivar connection_response: the status message returned by the server
    
    """
    
    def __init__(self, handler, path, func):
        """Initialize the instance.
        
        @type handler: L{HTTPCache}
        @param handler: the cache manager for the download
        @type path: C{list} of C{string}
        @param path: the server and path to download
        @type func: C{method}
        @param func: the method to call when the download completes
        
        """
        
        self.handler = handler
        self.path = path
        self.server = path[0]
        self.url = '/' + '/'.join(path[1:])
        self.func = func
        try:
            self.connection = HTTPConnection(self.server)
        except:
            logger.exception('cannot connect to http seed: '+self.server)
            return
        
        self.headers = {'User-Agent': VERSION}
        self.active = False
        logger.debug('starting thread to download '+self.url)
        rq = Thread(target = self._request, name = 'HTTPCache.CacheRequest._request')
        rq.setDaemon(False)
        rq.start()
        self.active = True

    def _request(self):
        """Do the request."""
        import encodings.ascii
        import encodings.punycode
        import encodings.idna
        
        try:
            logger.debug('sending request GET '+self.url+', '+str(self.headers))
            self.connection.request('GET',self.url, None, self.headers)
            
            r = self.connection.getresponse()
                
            logger.debug('got response '+str(r.status)+', '+r.reason+', '+str(r.getheaders()))
            self.connection_status = r.status
            self.connection_response = r.reason
            self.headers = dict(r.getheaders())
            self.received_data = r.read()
        except Exception, e:
            logger.exception('error accessing http server')
            self.connection_status = 500
            self.connection_response = 'Internal Server Error'
            self.headers = {}
            self.received_data = 'error accessing http server: '+str(e)
        self.handler.rawserver.add_task(self.request_finished)

    def request_finished(self):
        """Process the completed request."""
        self.connection.close()
        self.active = False
        self.handler.download_complete(self, self.path, self.func, 
                   (self.connection_status, self.connection_response, 
                    self.headers, self.received_data))


class HTTPCache:
    """Manage an HTTP download cache.
    
    @type rawserver: L{Debtorrent.RawServer.RawServer}
    @ivar rawserver: the server
    @type downloads: C{list} of L{CacheRequest}
    @ivar downloads: the list of all current downloads for the cache
    @type cachedir: C{string}
    @ivar cachedir: the directory to save cache files in
    
    """
    
    def __init__(self, rawserver, cachedir):
        """Initialize the instance.
        
        @type rawserver: L{Debtorrent.RawServer.RawServer}
        @param rawserver: the server
        @type cachedir: C{string}
        @param cachedir: the directory to save cache files in
        
        """
        
        self.rawserver = rawserver
        self.downloads = []
        self.cachedir = cachedir

    def download_get(self, path, func):
        """Create a new download from a site.
        
        @type path: C{list} of C{string}
        @param path: the server and path to download
        @type func: C{method}
        @param func: the method to call with the data when the download is complete
        
        """
        
        logger.info('Starting a downloader for: http://'+'/'.join(path))
        self.downloads.append(CacheRequest(self, path, func))

    def download_complete(self, d, path, func, r):
        """Remove a completed download from the list and process the data.
        
        Once a download has been completed, remove the downloader from the 
        list and save the downloaded file in the file system. Then return the
        data to the callback function. 
        
        @type d: L{CacheRequest}
        @param d: the cache request that is completed
        @type path: C{list} of C{string}
        @param path: the server and path that was downloaded
        @type func: C{method}
        @param func: the method to call with the data
        @type r: (C{int}, C{string}, C{dictionary}, C{string})
        @param r: the HTTP status code, status message, headers, and downloaded data
        
        """
        
        logger.info('download completed for: http://'+'/'.join(path))
        self.downloads.remove(d)

        file = self.get_filename(path)
        headers = {'Server': VERSION}

        if r[0] == 200:
            # Create the directory for the new file
            new_dir = split(file)[0]
            if new_dir != '' and not exists(new_dir):
                makedirs(new_dir)
            
            # Write the new file
            f = open(file, 'wb')
            f.write(r[3])
            f.close()
            
            # Set the modified time (on error use current time which should work)
            try:
                mtime = timegm(strptime(r[2]['last-modified'], time_format + ' %Z'))
                times = (mtime, mtime)
                utime(file, times)
            except:
                pass

        # Use the headers we want
        if exists(file):
            mtime_string = strftime(time_format + ' GMT', gmtime(getmtime(file)))
            headers['last-modified'] = mtime_string
        
        for k, v in r[2].items():
            if k in ('last-modified', 'content-type'):
                headers[k] = v
        
        # Call the callback function
        func(path, (r[0], r[1], headers, r[3]))

    def cache_get(self, path, uptodate = False, if_modified_time = ''):
        """Get the file from the cache.
        
        Will respond with the following HTTP status codes:
            - 200: the file was found in the cache and is up to date
            - 304: the file is up to date, but is not needed
            - 404: the file was not found in the cache
            - 405: the file was found, but is stale, and needs to be refreshed
        
        @type path: C{list} of C{string}
        @param path: the server and path to download
        @type uptodate: C{boolean}
        @param uptodate: whether to check the age of the file on the server to 
            see if the cached one is still current (optional, defaults to False)
        @type if_modified_time: C{string}
        @param if_modified_time: the if-modified-since header from the request
            (optional, defaults to not checking the if-modified-time)
        @rtype: (C{int}, C{string}, C{dictionary}, C{string})
        @return: the HTTP status code, status message, headers, and package data
        
        """
        
        file = self.get_filename(path)
            
        # Check if the file isn't in the cache
        if not exists(file):
            logger.info('cache miss: '+file)
            return (404, 'Not Found', {'Server': VERSION, 'Content-Type': 'text/plain', 'Pragma': 'no-cache'}, alas)
        
        if uptodate:
            # Get the last modified time from the server
            connection = HTTPConnection(path[0])
            connection.request('HEAD', '/' + '/'.join(path[1:]), None, {'User-Agent': VERSION})
            r = connection.getresponse()
            last_modified = r.getheader('last-modified')
            connection.close()

            # Check if the cached data is stale
            if self.check_mtime(last_modified, file = file) > 0:
                logger.info('cache out of date: '+file)
                return (405, 'Method Not Allowed', {'Server': VERSION, 'Content-Type': 'text/plain', 'Pragma': 'no-cache'}, alas)

        # Check if the request needs the data
        if if_modified_time and self.check_mtime(if_modified_time, file = file) >= 0:
            logger.info('cache up to date and so is request: '+file)
            return (304, 'Not Modified', {'Server': VERSION, 'Pragma': 'no-cache'}, '')
            
        # Read in the file and return the data
        f = open(file, 'rb')
        data = f.read()
        f.close()
        mtime_string = strftime(time_format+' GMT', gmtime(getmtime(file)))
        
        logger.info('cache hit: '+file)
        return (200, 'OK', {'Server': VERSION, 'Content-Type': 'text/plain', 'Last-Modified': mtime_string}, data)

    def get_filename(self, path):
        """Get the file name used for this path in the cache.
        
        @type path: C{list} of C{string}
        @param path: the server and path to download
        @rtype: C{string}
        @return: the file name
        
        """
        
        if not path:
            return None
        
        # Build the file name
        file = join(self.cachedir, path[0])
        for i in path[1:]:
            file = join(file, i)

        return file

    def get_file_mtime(self, path):
        """Get the modification time of the file in the cache.
        
        @type path: C{list} of C{string}
        @param path: the server and path to check
        @rtype: C{int}
        @return: the file's modification time, or 0 if the file is not found
        
        """
        
        if not path:
            return 0
        
        # Build the file name
        file = self.get_filename(path)

        try:
            return int(getmtime(file))
        except:
            return 0

    def check_mtime(self, http_mtime_string, path = [], file = '', server_mtime_string = ''):
        """Check the modified time of a file in the cache against a server header string.
        
        @type http_mtime_string: C{string}
        @param http_mtime_string: the modified time from an HTTP header
        @type path: C{list} of C{string}
        @param path: the server and path to download
            (optional, but one of file/path/server_mtime must be specified)
        @type file: C{string}
        @param file: the file name in the cache 
            (optional, but one of file/path/server_mtime must be specified)
        @type server_mtime_string: C{string}
        @param server_mtime_string: the last-modified time from the server's copy
            (optional, but one of file/path/server_mtime must be specified)
        @rtype: C{int}
        @return: the number of seconds the header's mtime is ahead of the 
            file's mtime (or None if the file isn't in the cache)
        
        """
        
        assert path or file or server_mtime_string

        if path:
            file = self.get_filename(path)
            
        if file:
            # Return None if the file isn't in the cache
            if not exists(file):
                return None
            
            # Check the server's time against the cached copy
            server_mtime = getmtime(file)
        else:
            server_mtime = timegm(strptime(server_mtime_string, time_format+' %Z'))

        http_mtime = timegm(strptime(http_mtime_string, time_format+' %Z'))

        return http_mtime - server_mtime
