# Written by Bram Cohen
# Modified by Cameron Dale
# see LICENSE.txt for license information
#
# $Id: HTTPHandler.py 199 2007-08-06 00:50:01Z camrdale-guest $

"""Handles incoming HTTP connections from other clients to this server.

@type logger: C{logging.Logger}
@var logger: the logger to send all log messages to for this module
@type weekdays: C{list} of C{string}
@var weekdays: the days of the week
@type months: C{list} of C{string}
@var months: the months of the year

"""

from cStringIO import StringIO
from sys import stdout, exc_info
import time
from clock import clock
from gzip import GzipFile
import signal, logging
try:
    True
except:
    True = 1
    False = 0

logger = logging.getLogger('DebTorrent.HTTPHandler')

weekdays = ['Mon', 'Tue', 'Wed', 'Thu', 'Fri', 'Sat', 'Sun']

months = [None, 'Jan', 'Feb', 'Mar', 'Apr', 'May', 'Jun',
    'Jul', 'Aug', 'Sep', 'Oct', 'Nov', 'Dec']

def isotime(secs = None):
    """Create an ISO formatted string of the time.
    
    @type secs: C{float}
    @param secs: number of seconds since the epoch 
        (optional, default is to use the current time)
    @rtype: C{string}
    @return: the ISO formatted string representation of the time
    
    """
    
    if secs == None:
        secs = time.time()
    return time.strftime('%Y-%m-%d %H:%M UTC', time.gmtime(secs))

class HTTPConnection:
    """A single connection from an HTTP client.
    
    Handles a single HTTP GET or HEAD request from a client using HTTP/1.0.
    
    @type handler: L{HTTPHandler}
    @ivar handler: the handler for all incoming HTTP connections
    @type connection: L{SocketHandler.SingleSocket}
    @ivar connection: the new connection that was created
    @type buf: C{string}
    @ivar buf: the buffered data received on the connection
    @type closed: C{boolean}
    @ivar closed: whether the connection has been closed
    @type done: C{boolean}
    @ivar done: whether the response has been sent
    @type donereading: C{boolean}
    @ivar donereading: whether the headers have been read
    @type next_func: C{method}
    @ivar next_func: the next method to call to process data on this connection
    @type header: C{string}
    @ivar header: the first header line received from the request
    @type command: C{string}
    @ivar command: the requested command ('GET' or 'HEAD')
    @type pre1: C{boolean}
    @ivar pre1: whether the request is from a pre version 1.0 client
    @type headers: C{dictionary}
    @ivar headers: the headers received with the request
    @type encoding: C{string}
    @ivar encoding: the encoding to use when sending the response
    
    """
    
    def __init__(self, handler, connection):
        """Initialize the instance.
        
        @type handler: L{HTTPHandler}
        @param handler: the handler for all incoming HTTP connections
        @type connection: L{SocketHandler.SingleSocket}
        @param connection: the new connection that was created
        
        """
        
        self.handler = handler
        self.connection = connection
        self.buf = ''
        self.closed = False
        self.done = False
        self.donereading = False
        self.next_func = self.read_type

    def get_ip(self):
        """Get the IP address of the connection.
        
        @rtype: C{string}
        @return: the IP address
        
        """
        
        return self.connection.get_ip()

    def data_came_in(self, data):
        """Process data that came in on the connection.
        
        Processes the request lines one at a time by calling the L{next_func}
        method with the lines.
        
        @type data: C{string}
        @param data: the data that came in
        @rtype: C{boolean}
        @return: whether to keep the connection open
        
        """
        
        if self.donereading or self.next_func is None:
            return True
        self.buf += data
        while True:
            try:
                i = self.buf.index('\n')
            except ValueError:
                return True
            val = self.buf[:i]
            self.buf = self.buf[i+1:]
            self.next_func = self.next_func(val)
            if self.donereading:
                return True
            if self.next_func is None or self.closed:
                return False

    def read_type(self, data):
        """Process the first header line that came in.
        
        @type data: C{string}
        @param data: the line that came in
        @rtype: C{method}
        @return: the next method to call to process data on this connection
        
        """
        
        self.header = data.strip()
        words = data.split()
        if len(words) == 3:
            self.command, self.path, garbage = words
            self.pre1 = False
        elif len(words) == 2:
            self.command, self.path = words
            self.pre1 = True
            if self.command != 'GET':
                logger.warning('connection closed, improper command: '+self.command)
                return None
        else:
            logger.warning('connection closed, corrupt header line: '+data)
            return None
        if self.command not in ('HEAD', 'GET'):
            logger.warning('connection closed, improper command: '+self.command)
            return None
        self.headers = {}
        return self.read_header

    def read_header(self, data):
        """Process the next header line that came in.
        
        @type data: C{string}
        @param data: the line that came in
        @rtype: C{method}
        @return: the next method to call to process data on this connection
        
        """
        
        data = data.strip()
        if data == '':
            self.donereading = True
            if self.headers.get('accept-encoding','').find('gzip') > -1:
                self.encoding = 'gzip'
            else:
                self.encoding = 'identity'
            r = self.handler.getfunc(self, self.path, self.headers)
            if r is not None:
                self.answer(r)
            return None
        try:
            i = data.index(':')
        except ValueError:
            logger.warning('connection closed, corrupt header: '+data)
            return None
        self.headers[data[:i].strip().lower()] = data[i+1:].strip()
        logger.debug(data[:i].strip() + ": " + data[i+1:].strip())
        return self.read_header

    def answer(self, (responsecode, responsestring, headers, data)):
        """Send a response to the client on the connection and close it.
        
        @type responsecode: C{int}
        @param responsecode: the response code to send
        @type responsestring: C{string}
        @param responsestring: the response string to send
        @type headers: C{dictionary}
        @param headers: the headers to send with the response
        @type data: C{string}
        @param data: the data to send with the response
        
        """
        
        if self.closed:
            logger.warning('connection closed before anwswer, dropping data')
            return
        if self.encoding == 'gzip':
            compressed = StringIO()
            gz = GzipFile(fileobj = compressed, mode = 'wb', compresslevel = 9)
            gz.write(data)
            gz.close()
            cdata = compressed.getvalue()
            if len(cdata) >= len(data):
                self.encoding = 'identity'
            else:
                logger.debug('Compressed: '+str(len(cdata))+'  Uncompressed: '+str(len(data)))
                data = cdata
                headers['Content-Encoding'] = 'gzip'

        # i'm abusing the identd field here, but this should be ok
        if self.encoding == 'identity':
            ident = '-'
        else:
            ident = self.encoding
        self.handler.write_log( self.connection.get_ip(), ident, '-',
                                self.header, responsecode, len(data),
                                self.headers.get('referer','-'),
                                self.headers.get('user-agent','-') )
        self.done = True
        logger.info('sending response: '+str(responsecode)+' '+responsestring+
                    ' ('+str(len(data))+' bytes)')
        r = StringIO()
        r.write('HTTP/1.0 ' + str(responsecode) + ' ' + 
            responsestring + '\r\n')
        if not self.pre1:
            headers['Content-Length'] = len(data)
            for key, value in headers.items():
                r.write(key + ': ' + str(value) + '\r\n')
            r.write('\r\n')
        if self.command != 'HEAD':
            r.write(data)
        self.connection.write(r.getvalue())
        if self.connection.is_flushed():
            self.connection.shutdown(1)

class HTTPHandler:
    """The handler for all new and existing HTTP connections.
    
    @type connections: C{dictionary}
    @ivar connections: all the existing connections, keys are the connection 
        objects, values are the L{HTTPConnection} objects
    @type getfunc: C{method}
    @ivar getfunc: the method to call with the processed GET requests
    @type minflush: C{float}
    @ivar minflush: the minimum amount of time between flushing the log
    @type lastflush: C{float}
    @ivar lastflush: the time of the last log flush
    @type logfile: C{string}
    @ivar logfile: the file name to write the logs to
    @type log: C{file}
    @ivar log: the file to write the logs to
    
    """
    
    def __init__(self, getfunc, minflush, logfile = None, hupmonitor = None):
        """Initialize the instance.
        
        @type getfunc: C{method}
        @param getfunc: the method to call with the processed GET requests
        @type minflush: C{float}
        @param minflush: the minimum amount of time between flushing the log
        @type logfile: C{string}
        @param logfile: the file to write the logs to
            (optional, defaults to standard output)
        @type hupmonitor: C{boolean}
        @param hupmonitor: whether to reopen the log file on a HUP signal
            (optional, default is False)
        
        """
        
        self.connections = {}
        self.getfunc = getfunc
        self.minflush = minflush
        self.lastflush = clock()
        self.logfile = None
        self.log = None
        if (logfile) and (logfile != '-'):
            try:
                self.logfile = logfile
                self.log = open(self.logfile,'a')
                print >> self.log, "# Log Started: ", isotime()
            except:
                logger.exception('could not open log file')
            else:
                if hupmonitor:
                    def huphandler(signum, frame, self = self):
                        try:
                            self.log.close()
                            self.log = open(self.logfile,'a')
                            print >> self.log, "# Log reopened: ", isotime()
                        except:
                            logger.exception('could not reopen log file')
                     
                    signal.signal(signal.SIGHUP, huphandler)            

    def external_connection_made(self, connection):
        """Create a new HTTPConnection object.
        
        @type connection: L{SocketHandler.SingleSocket}
        @param connection: the connection
        
        """
        
        self.connections[connection] = HTTPConnection(self, connection)

    def connection_flushed(self, connection):
        """Flush a connection.
        
        @type connection: L{SocketHandler.SingleSocket}
        @param connection: the connection
        
        """
        
        if self.connections[connection].done:
            connection.shutdown(1)

    def connection_lost(self, connection):
        """Remove the lost connection from the existing ones.
        
        @type connection: L{SocketHandler.SingleSocket}
        @param connection: the connection
        
        """
        
        ec = self.connections[connection]
        ec.closed = True
        del ec.connection
        del ec.next_func
        del self.connections[connection]

    def data_came_in(self, connection, data):
        """Process incoming data on an existing connection.
        
        @type connection: L{SocketHandler.SingleSocket}
        @param connection: the connection
        @type data: C{string}
        @param data: the data that came in
        
        """
        
        c = self.connections[connection]
        if not c.data_came_in(data) and not c.closed:
            c.connection.shutdown(1)

    def write_log(self, ip, ident, username, header,
            responsecode, length, referrer, useragent):
        """Print a log message.
        
        @type ip: C{string}
        @param ip: the IP address of the connection
        @type ident: C{string}
        @param ident: the type of connection
        @type username: C{string}
        @param username: the username from the client
        @type header: C{string}
        @param header: the first header line of the request
        @type responsecode: C{int}
        @param responsecode: the HTTP status code of the response
        @type length: C{int}
        @param length: the number of bytes in the data returned
        @type referrer: C{string}
        @param referrer: the referer for the incoming request
        @type useragent: C{string}
        @param useragent: the client's useragent used to make the request
        
        """
        
        year, month, day, hour, minute, second, a, b, c = time.localtime(time.time())
        print >> self.log, '%s %s %s [%02d/%3s/%04d:%02d:%02d:%02d] "%s" %i %i "%s" "%s"' % (
            ip, ident, username, day, months[month], year, hour,
            minute, second, header, responsecode, length, referrer, useragent)
        t = clock()
        if t - self.lastflush > self.minflush:
            self.lastflush = t
            if self.log:
                self.log.flush()
            else:
                stdout.flush()
